@@ -4,7 +4,7 @@
 #
 # timsieved client script
 #
-# Copyright © 2006-2008 Phil Pennock.  All rights reserved.
+# Copyright © 2006-2009 Phil Pennock.  All rights reserved.
 #
 # Redistribution and use in source and binary forms, with or without
 # modification, are permitted provided that the following conditions
@@ -48,7 +48,7 @@ my @cmd_localfs_ls = qw( ls -C );
 # No user-serviceable parts below
 
 our $VERSION = 0;
-my $svn_revision = '$Revision: 58 $';
+my $svn_revision = '$Revision: 69 $';
 if ($svn_revision =~ /^.Revision:\s*(\d+)\s*\$\z/) {
 	$svn_revision = $1;
 	$VERSION = '0.' . $1;
@@ -78,6 +78,39 @@ use POSIX qw/ strftime /;
 use Term::ReadKey;
 # interactive mode will attempt to pull in Term::ReadLine too.
 
+my $DEBUGGING = 0;
+
+sub do_version_display {
+	print "${0}: Version $VERSION\n";
+	if ($DEBUGGING) {
+		my @do_require = (
+			'Authen::SASL::Perl::GSSAPI',
+			'Term::ReadLine',
+		);
+		foreach my $r (@do_require) {
+			(my $rr = $r) =~ s,::,/,g;
+			eval { require "$rr.pm" };
+		}
+		foreach my $mod (
+				'Authen::SASL',
+				'Authen::SASL::Perl',
+				'IO::Socket::INET6',
+				'IO::Socket::SSL',
+				'Term::ReadKey',
+				@do_require) {
+			my $vname = "${mod}::VERSION";
+			my $ver;
+			eval { no strict 'refs'; $ver = ${$vname} };
+			if (defined $ver) {
+				print "  Module $mod Version $ver\n";
+			} else {
+				print "  Module $mod -- no version number available\n";
+			}
+		}
+	}
+	exit 0;
+}
+
 sub debug;
 sub sent;
 sub ssend;
@@ -88,13 +121,14 @@ sub closedie;
 sub closedie_NOmsg;
 sub die_NOmsg;
 
-my $DEBUGGING = 0;
+my $DEBUGGING_SASL = 0;
 my $DATASTART = tell DATA;
 my $localsievename;
 my $remotesievename;
 my ($user, $authzid, $authmech, $sslkeyfile, $sslcertfile, $passwordfd);
 my $prioritise_auth_external = 0;
 my $dump_tls_information = 0;
+my $opt_version_req = 0;
 my ($server, $realm);
 my $port = 'sieve(2000)';
 my $net_domain = AF_UNSPEC;
@@ -118,6 +152,7 @@ GetOptions(
 	"4"		=> sub { $net_domain = AF_INET },
 	"6"		=> sub { $net_domain = AF_INET6 },
 	"debug"		=> \$DEBUGGING,
+	"debugsasl"	=> \$DEBUGGING_SASL,
 	"dumptlsinfo|dumpsslinfo"	=> \$dump_tls_information,
 	# option names can be short-circuited, $action is complete:
 	"upload"	=> sub { $action = 'upload' },
@@ -129,10 +164,12 @@ GetOptions(
 	"exec|e=s"	=> sub { $execscript = $_[1]; $action='command-loop' },
 	'help|?'	=> sub { pod2usage(0) },
 	'man'		=> sub { pod2usage(-exitstatus => 0, -verbose => 2) },
-	'version'	=> sub { print "${0}: Version $VERSION\n"; exit 0 },
+	'version'	=> \$opt_version_req, # --version --debug should work
 ) or pod2usage(2);
 # We don't implement HAVESPACE <script> <size>
 
+do_version_display() if $opt_version_req;
+
 if (defined $ARGV[0] and not defined $server) {
 	# sieveshell compatibility.
 	my $where = $ARGV[0];
@@ -247,7 +284,8 @@ my %capa_permit_empty = (
 	SASL	=> sub {
 		return if exists $_[1]{STARTTLS};
 		# We die because there's no way to authenticate.
-		# Spec states that after STARTTLS SASL must be non-empty
+		# Spec states "This list can be empty if and only if STARTTLS
+		# is also advertised" (section 1.7).
 		closedie $_[0], "Empty SASL not permitted without STARTTLS\n";
 		},
 	SIEVE	=> undef,
@@ -257,6 +295,7 @@ sub parse_capabilities
 {
 	my $sock = shift;
 	local %_ = @_;
+	# Used under TLS to coerce EXTERNAL auth to be preferred:
 	my $external_first = 0;
 	$external_first = $_{external_first} if exists $_{external_first};
 
@@ -264,10 +303,49 @@ sub parse_capabilities
 	%raw_capabilities = ();
 	%capa = ();
 	while (<$sock>) {
-		received;
+		received unless /^OK\b/;
 		chomp; s/\s*$//;
 		if (/^OK\b/) {
-			last unless exists $_{until_see_no};
+			sget($sock, '-firstline', $_);
+			last unless exists $_{sent_a_noop};
+			# See large comment below in STARTTLS explaining the
+			# resync problem to understand why this is here.
+			my $end_tag = $_{sent_a_noop};
+			unless (defined $end_tag and length $end_tag) {
+				# In the initial NOOP-featuring draft, #10, we
+				# got back 'NOOP'.  However, this was at odds
+				# with the general syntax rules, so #11/#12
+				# added the TAG response; with this, the
+				# supplied NOOP parameter is returned in the
+				# TAG response, but if there's no parameter
+				# then there's just arbitrary server text.
+				#
+				# So where this used to use a default $end_tag
+				# of 'NOOP', now we declare it a coding error
+				# for this script to pass sent_a_noop without
+				# a value consisting of the tag.
+				closedie $sock, "Internal error: sent_a_noop without tag\n";
+			}
+			# Play crude, just look for the tag anywhere in the
+			# response, honouring only word boundaries.  It's our
+			# responsibility to make the tag long enough that this
+			# works without tokenising.
+			# Really, should check for: OK (TAG <tag-string>) text
+			# where <tag-string> is "$end_tag" or {<len>}\r\n$end_tag
+			if ($_ =~ m/\b\Q${end_tag}\E\b/) {
+				return;
+			}
+			# Okay, that's the "server understands NOOP" case, for
+			# which the server should have advertised the
+			# capability prior to TLS (and so subject to
+			# tampering); we play fast and loose, sending NOOP in
+			# all cases, so have to cover the NO case below too;
+			# the known instance of protocol violation we know of
+			# is an older server waiting for client command after
+			# TLS is up.  That server doesn't support NOOP.
+			# Sending NOOP and expecting a NO response for the
+			# unsupported command was the original technique used
+			# by this code.
 		} elsif (/^\"([^"]+)\"\s+\"(.*)\"$/) {
 			my ($k, $v) = (uc($1), $2);
 			unless (length $v) {
@@ -300,13 +378,13 @@ sub parse_capabilities
 			$raw_capabilities{$1} = '';
 			$capa{$1} = 1;
 		} elsif (/^NO\b/) {
-			return if exists $_{until_see_no};
-			warn "Unhandled server line: $_\n"
+			return if exists $_{sent_a_noop};
+			warn "Unhandled server line: $_\n";
 		} elsif (/^BYE\b(.*)/) {
 			closedie_NOmsg $sock, $1,
 				"Server said BYE when we expected capabilities.\n";
 		} else {
-			warn "Unhandled server line: $_\n"
+			warn "Unhandled server line: $_\n";
 		}
 	};
 
@@ -349,6 +427,8 @@ sub parse_capabilities
 }
 parse_capabilities $sock;
 
+my $tls_bitlength = -1;
+
 if (exists $capa{STARTTLS}) {
 	ssend $sock, "STARTTLS";
 	sget $sock;
@@ -357,7 +437,11 @@ if (exists $capa{STARTTLS}) {
 		my $e = IO::Socket::SSL::errstr();
 		die "STARTTLS promotion failed: $e\n";
 	};
-	debug("--- TLS activated here");
+	if (exists $main::{"Net::"} and exists $main::{"Net::"}{"SSLeay::"}) {
+		my $t = Net::SSLeay::get_cipher_bits($sock->_get_ssl_object(), 0);
+		$tls_bitlength = $t if defined $t and $t;
+	}
+	debug("--- TLS activated here [$tls_bitlength bits]");
 	if ($dump_tls_information) {
 		print $sock->dump_peer_certificate();
 		if ($DEBUGGING and
@@ -381,15 +465,42 @@ if (exists $capa{STARTTLS}) {
 	# reality, so that we can get back to command-response.
 	# We can't just check to see if there's data to read or not, since
 	# that will break if the next data is delayed (race condition).
-	# There is no protocol-compliant method to determine this, short
+	# There was no protocol-compliant method to determine this, short
 	# of "wait a while, see if anything comes along; if not, send
-	# CAPABILITY ourselves".  So, I break protocol by sending the
+	# CAPABILITY ourselves".  So, I broke protocol by sending the
 	# non-existent command NOOP, then scan for the resulting NO.
-	# This at least is stably deterministic.
-	ssend $sock, "NOOP";
-	parse_capabilities($sock,
-		until_see_no	=> 1,
-		external_first => $prioritise_auth_external);
+	# This at least is stably deterministic.  However, from draft 10
+	# onwards, NOOP is a registered available extension which returns
+	# OK.
+	#
+	# New problem: again, Cyrus timsieved.  As of 2.3.13, it drops the
+	# connection for an unknown command instead of returning NO.  And
+	# logs "Lost connection to client -- exiting" which is an interesting
+	# way of saying "we dropped the connection".  At this point, I give up
+	# on protocol-deterministic checks and fall back to version checking.
+	# Alas, Cyrus 2.2.x is still widely deployed because 2.3.x is the
+	# development series and 2.2.x is officially the stable series.
+	# This means that if they don't support NOOP by 2.3.14, I have to
+	# figure out how to decide what is safe and backtrack which version
+	# precisely was the first to send the capability response correctly.
+	my $use_noop = 1;
+	if (exists $capa{"IMPLEMENTATION"} and
+		$capa{"IMPLEMENTATION"} =~ /^Cyrus timsieved v2\.3\.(\d+)\z/ and
+		$1 >= 13) {
+		debug("--- Cyrus drops connection with dubious log msg if send NOOP, skip that");
+		$use_noop = 0;
+	}
+
+	if ($use_noop) {
+		my $noop_tag = "STARTTLS-RESYNC-CAPA";
+		ssend $sock, qq{NOOP "$noop_tag"};
+		parse_capabilities($sock,
+			sent_a_noop	=> $noop_tag,
+			external_first	=> $prioritise_auth_external);
+	} else {
+		parse_capabilities($sock,
+			external_first	=> $prioritise_auth_external);
+	}
 	unless (scalar keys %capa) {
 		ssend $sock, "CAPABILITY";
 		parse_capabilities($sock,
@@ -400,6 +511,9 @@ if (exists $capa{STARTTLS}) {
 }
 
 my %authen_sasl_params;
+if ($DEBUGGING_SASL) {
+	$authen_sasl_params{debug} = 15;
+}
 $authen_sasl_params{callback}{user} = $user;
 if (defined $authzid) {
 	$authen_sasl_params{callback}{authname} = $authzid;
@@ -450,6 +564,9 @@ my $secflags = 'noanonymous';
 $secflags .= ' noplaintext' if $forbid_clearauth;
 my $authconversation = $sasl->client_new('sieve', $server, $secflags)
 	or die "SASL conversation init failed (local problem): $!\n";
+if ($tls_bitlength > 0) {
+	$authconversation->property(externalssf => $tls_bitlength);
+}
 if (defined $realm) {
 	$authconversation->property(realm => $realm);
 }
@@ -750,13 +867,14 @@ if (defined $execscript) {
 				sub { complete_rl_sieve($term, $sock, @_) };
 			$term->Attribs->{completer_quote_characters} = '"';
 			$term->Attribs->{filename_quote_characters} = " \t";
+			$term->call_function('display-readline-version') if $DEBUGGING;
 		}
 		$cmdlineget_func = sub { return $term->readline('> ') };
 		print STDERR "ReadLine support enabled.\n";
 	};
 	unless (defined $cmdlineget_func) {
 		$cmdlineget_func = sub {
-			print "> "; $| = 1;
+			print "> "; local $| = 1;
 			my $l = <STDIN>;
 			return $l;
 		};
@@ -1234,10 +1352,22 @@ sub ssend
 sub sget
 {
 	my $sock = shift;
+	my $l = undef;
 	my $dochomp = 1;
-	$dochomp = 0 if defined $_[0] and $_[0] eq '-nochomp';
-	my $l;
-	$l = $sock->getline();
+	while (@_) {
+		my $t = shift;
+		next unless defined $t;
+		if ($t eq '-nochomp') { $dochomp = 0; next; }
+		if ($t eq '-firstline') {
+			die "Missing sget -firstline parameter"
+				unless defined $_[0];
+			$l = $_[0];
+			shift;
+			next;
+		}
+		die "Unknown sget parameter [$t]";
+	}
+	$l = $sock->getline() unless defined $l;
 	unless (defined $l) {
 		debug "... no line read, connection dropped?";
 		die "Connection dropped unexpectedly when trying to read.\n";
@@ -1403,10 +1533,14 @@ people think about names in the local filesystem.  There is no default
 script name.
 
 The B<--debug> option turns on diagnostic traces.
+The B<--debugsasl> option asks the SASL library for debugging.
 The B<--dumptlsinfo> shows the TLS (SSL) peer information; if specified
 together with B<--debug> then the server's PEM certificate will be
 provided as debug trace.
 
+The B<--version> option shows version information.
+When combined with B<--debug> it will show implementation dependency versions.
+
 The server can be a host or IP address, IPv4 or IPv6;
 the default is C<$IMAP_SERVER> from the environment (if it's not a
 unix-domain socket path) with any port specificaion stripped off,
